#include <stdafx.h>

#include "MBitmap.h"


CMBitmap::CMBitmap()
{
	// MG: Initialize our only non-class data member
	memset( &TheBM, 0, sizeof(TheBM) );
}

// MG: Standard Serialize support
void CMBitmap::Serialize( CArchive& ar )
{
	if( ar.IsStoring() )
	{
		if( !SaveBitmap( ar.GetFile() ) )
			::MessageBox( NULL, "Error saving bitmap", NULL, MB_OK );
	}
	else
	{
		if( !LoadBitmap( ar.GetFile() ) )
			::MessageBox( NULL, "Error loading bitmap", NULL, MB_OK );
	}
}


// MG: Load bitmap from a CFile - Real work takes place here
BOOL CMBitmap::LoadBitmap( CFile* BMPFile)
{
    // MG: Wipe out any data we have
	DeleteObject();
	
	// MG: Read and validate BMP file header
	BITMAPFILEHEADER bmfHeader;
	if (BMPFile->Read((LPSTR)&bmfHeader, sizeof(bmfHeader)) != sizeof(bmfHeader))
		return FALSE;
    if (bmfHeader.bfType != 'MB' )
		return FALSE;

	// MG: Determine data size need for BMP, and allocate
	DWORD nDataSize;
	nDataSize = bmfHeader.bfSize - sizeof(BITMAPFILEHEADER);
	HGLOBAL hDIB = ::GlobalAlloc(GMEM_FIXED, nDataSize);
	if (hDIB == 0)
		return FALSE;
	
	// MG: Read BMP data and validate length
	if (BMPFile->Read((LPSTR)hDIB, nDataSize) != nDataSize )
    {
		::GlobalFree(hDIB);
		return FALSE;
	}

	// MG: Start setting up structure data
	BITMAPINFOHEADER &bmiHeader = *(LPBITMAPINFOHEADER)hDIB ;
    BITMAPINFO &bmInfo = *(LPBITMAPINFO)hDIB ;

	int nColors = bmiHeader.biClrUsed ? bmiHeader.biClrUsed : 1 << bmiHeader.biBitCount;

    // MG: Setup pointer into Bitmap data
	LPVOID lpData;
    if( bmInfo.bmiHeader.biBitCount > 8 )
		lpData = (LPVOID)((LPDWORD)(bmInfo.bmiColors + bmInfo.bmiHeader.biClrUsed) + 
                ((bmInfo.bmiHeader.biCompression == BI_BITFIELDS) ? 3 : 0));
    else
		lpData = (LPVOID)(bmInfo.bmiColors + nColors);

	// MG Create and init the Windows palette
	if( nColors <= 256 )
	{
		UINT nSize = sizeof(LOGPALETTE) + (sizeof(PALETTEENTRY) * nColors);
		LOGPALETTE *pLP = (LOGPALETTE *) GlobalAlloc( GMEM_FIXED, nSize );
		if( !pLP )
		{
			::GlobalFree(hDIB);
			return FALSE;
		}
		pLP->palVersion = 0x300;
		pLP->palNumEntries = nColors;
		for( int i=0; i < nColors; i++)
		{
			pLP->palPalEntry[i].peRed = bmInfo.bmiColors[i].rgbRed;
			pLP->palPalEntry[i].peGreen = bmInfo.bmiColors[i].rgbGreen;
			pLP->palPalEntry[i].peBlue = bmInfo.bmiColors[i].rgbBlue;
			pLP->palPalEntry[i].peFlags = 0;
		}
		Pal.CreatePalette( pLP );
		GlobalFree( pLP );
	}
	
	CClientDC dc(NULL);

	CPalette* pOldPalette = NULL;
	if( nColors <= 256 )
	{
		pOldPalette = dc.SelectPalette( &Pal, FALSE );
		dc.RealizePalette();
	}

	// MG: Create the Windows HBITMAP
	HBITMAP hBmp = CreateDIBitmap( dc.m_hDC,&bmiHeader,CBM_INIT,lpData,
		&bmInfo,DIB_RGB_COLORS);

	// MG: Attach the Windows HBITMAP to our object (A CBitmap function)
	Attach( hBmp );
	if( pOldPalette )
		dc.SelectPalette( pOldPalette, FALSE );

	// MG: Initialize our data member
	InitBM(TRUE);

	::GlobalFree(hDIB);
	return TRUE;

}

BOOL CMBitmap::LoadBitmap( char* lpszFileName )
{
	CFile BMPFile;
	if( !BMPFile.Open( lpszFileName, CFile::modeRead) )
		return FALSE;
	
    return( LoadBitmap( &BMPFile ) );
}

// MG: DeleteObject 'clears-out' a Bitmap.  Under MFC, you can't load a CBitmap twice
// without first releasing it.  This routine is called automatically by all of the 
// CMBitmap loading routines
void CMBitmap::DeleteObject()
{
	if( m_hObject )
		CBitmap::DeleteObject();
	if( Pal.m_hObject )
		Pal.DeleteObject();
	memset( &TheBM, 0, sizeof(TheBM) );

}

BOOL CMBitmap::SaveBitmap( CFile* BMPFile )
{
	// MG: Much of this code was borrowed from the online help, but I found that
	// there were bugs in the example code, specifically where it dealt with the number
	// of colors (It used the Bitsperpixel value as if it were number of colors).

	BITMAPFILEHEADER hdr;       /* bitmap file-header */ 
    LPBYTE lpBits;              /* memory pointer */ 
    BITMAPINFO* pbmi;
	BITMAPINFOHEADER* pbmih;
    WORD    cClrBits; 
 
    cClrBits = (WORD)(TheBM.bmPlanes * TheBM.bmBitsPixel); 
 
    if (cClrBits == 1)		 cClrBits = 1; 
    else if (cClrBits <= 4)  cClrBits = 4; 
    else if (cClrBits <= 8)  cClrBits = 8; 
    else if (cClrBits <= 16) cClrBits = 16; 
    else if (cClrBits <= 24) cClrBits = 24; 
    else cClrBits = 32; 
 
    if (cClrBits < 24) 
         pbmi = (PBITMAPINFO) GlobalAlloc(GMEM_FIXED, 
                    sizeof(BITMAPINFOHEADER) + 
                    sizeof(RGBQUAD) * (1<<cClrBits)); 
    else 
         pbmi = (PBITMAPINFO) GlobalAlloc(GMEM_FIXED, sizeof(BITMAPINFOHEADER)); 

	pbmih = (BITMAPINFOHEADER*)pbmi;
    pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER); 
    pbmi->bmiHeader.biWidth = TheBM.bmWidth; 
    pbmi->bmiHeader.biHeight = TheBM.bmHeight; 
    pbmi->bmiHeader.biPlanes = TheBM.bmPlanes; 
    pbmi->bmiHeader.biBitCount = TheBM.bmBitsPixel; 
    if (cClrBits < 24) 
        pbmi->bmiHeader.biClrUsed = 1<<cClrBits; 
	pbmi->bmiHeader.biCompression = BI_RGB; 
    pbmi->bmiHeader.biSizeImage = (pbmi->bmiHeader.biWidth + 7) /8 
                                  * pbmi->bmiHeader.biHeight 
                                  * cClrBits; 
    pbmi->bmiHeader.biClrImportant = 0; 

	// We now have a BitmapInfoHeader

    lpBits = (LPBYTE) GlobalAlloc(GMEM_FIXED, pbmih->biSizeImage);
    
 	CClientDC dc(NULL);
   
	if(!GetDIBits(dc.m_hDC, (HBITMAP)m_hObject, 0, (WORD) pbmih->biHeight, lpBits, (BITMAPINFO*)pbmih, DIB_RGB_COLORS) )
	{
		MessageBox( NULL, "Failed", NULL, MB_OK );
		int i=GetLastError();
		return(FALSE);
	}
        
 
    hdr.bfType = 'MB';
 
    /* Compute the size of the entire file. */ 
	if( cClrBits < 24 )
		hdr.bfSize = (DWORD) (sizeof(BITMAPFILEHEADER) + 
			         pbmih->biSize + (1<<cClrBits) 
				     * sizeof(RGBQUAD) + pbmih->biSizeImage);
	else
		hdr.bfSize = (DWORD) (sizeof(BITMAPFILEHEADER) + 
			         pbmih->biSize + pbmih->biSizeImage);
 
    hdr.bfReserved1 = 0; 
    hdr.bfReserved2 = 0; 
 
    /* Compute the offset to the array of color indices. */ 
	hdr.bfOffBits = (DWORD) sizeof(BITMAPFILEHEADER) + 
                    pbmih->biSize + (1<<cClrBits) 
                    * sizeof (RGBQUAD); 
 
    BMPFile->Write( &hdr, sizeof(BITMAPFILEHEADER) );
    BMPFile->Write( pbmi, sizeof(BITMAPINFOHEADER) + (1<<cClrBits) * sizeof (RGBQUAD) );
    BMPFile->Write( lpBits, pbmih->biSizeImage );

    /* Free memory. */
    GlobalFree((HGLOBAL)lpBits);
	GlobalFree((HGLOBAL)pbmi );
	return(TRUE);
}

BOOL CMBitmap::SaveBitmap( char* lpszFileName )
{
	CFile BMPFile;

	if( !BMPFile.Open( lpszFileName, CFile::modeWrite|CFile::modeCreate) )
		return FALSE;
	return( SaveBitmap( &BMPFile ) );
}

void CMBitmap::Draw( CDC*pDC, int X, int Y, int Width, int Height, int SrcX, int SrcY, int Mode )
{
	if( !m_hObject )
		return;
	CDC memDC;

	// Create a memory DC
	memDC.CreateCompatibleDC( pDC );
	memDC.SelectObject( this );

	// Select and realize the palette
	if( (pDC->GetDeviceCaps(RASTERCAPS) & RC_PALETTE) && Pal.m_hObject )
	{
		pDC->SelectPalette( &Pal, FALSE );
		pDC->RealizePalette();
	}
	if( !Width )
		Width = TheBM.bmWidth;
	if( !Height )
		Height = TheBM.bmHeight;
    pDC->BitBlt(X, Y, Width, Height, &memDC, SrcX, SrcY, Mode );
}

void CMBitmap::Stretch(CDC*pDC, int X, int Y, int nWidth, int nHeight, int xSrc, int ySrc, int nSrcWidth, int nSrcHeight, DWORD dwRop )
{
	if( !m_hObject )
		return;
	CDC memDC;

	// Create a memory DC
	memDC.CreateCompatibleDC( pDC );
	memDC.SelectObject( this );

	// Select and realize the palette
	if( (pDC->GetDeviceCaps(RASTERCAPS) & RC_PALETTE) && Pal.m_hObject )
	{
		pDC->SelectPalette( &Pal, FALSE );
		pDC->RealizePalette();
	}
	if( !nWidth )
		nWidth = TheBM.bmWidth;
	if( !nHeight )
		nHeight = TheBM.bmHeight;
	if( !nSrcWidth )
		nSrcWidth = TheBM.bmWidth;
	if( !nSrcHeight )
		nSrcHeight= TheBM.bmHeight;
    pDC->StretchBlt(X, Y, nWidth, nHeight, &memDC, xSrc, ySrc, nSrcWidth, nSrcHeight, dwRop );
}

// MG: Initialize TheBM and bitmap dimensions
BOOL CMBitmap::InitBM(BOOL OK) 
{ 
	if(OK)
		GetObject( sizeof(TheBM), &TheBM ); 
	else 
		memset(&TheBM,0,sizeof(TheBM)); 
	SetBitmapDimension( TheBM.bmWidth, TheBM.bmHeight );
	return(OK);
}
	
// MG: The following routines are the standard CBitmap loading routines, but we want to make
// sure that we do DeleteObject before each call, and InitBM after each.
BOOL CMBitmap::LoadBitmap( UINT X ) 
{ 
	DeleteObject(); 
	return( InitBM(CBitmap::LoadBitmap( X ) ) ); 
}
BOOL CMBitmap::LoadOEMBitmap( UINT nIDBitmap ) 
{ 
	DeleteObject(); 
	return(InitBM(CBitmap::LoadOEMBitmap( nIDBitmap ))); 
}
BOOL CMBitmap::LoadMappedBitmap( UINT nIDBitmap, UINT nFlags, LPCOLORMAP lpColorMap, int nMapSize )
{
	DeleteObject();
	return( InitBM(CBitmap::LoadMappedBitmap( nIDBitmap, nFlags, lpColorMap, nMapSize ) ) );
}
BOOL CMBitmap::CreateBitmap( int nWidth, int nHeight, UINT nPlanes, UINT nBitcount, const void* lpBits )
{
	DeleteObject();
	return( InitBM(CBitmap::CreateBitmap( nWidth, nHeight, nPlanes, nBitcount, lpBits ) ) );
}
BOOL CMBitmap::CreateBitmapIndirect( LPBITMAP lpBitmap ) 
{ 
	DeleteObject(); 
	return( InitBM(CBitmap::CreateBitmapIndirect( lpBitmap ) ) ); 
}
BOOL CMBitmap::CreateCompatibleBitmap( CDC* pDC, int nWidth, int nHeight ) 
{
	DeleteObject(); 
	return( InitBM(CBitmap::CreateCompatibleBitmap( pDC, nWidth, nHeight ) ) ); 
	
}
BOOL CMBitmap::CreateDiscardableBitmap( CDC* pDC, int nWidth, int nHeight ) 
{
	DeleteObject(); 
	return( InitBM(CBitmap::CreateDiscardableBitmap( pDC, nWidth, nHeight ) ) ); 
}
